#include <sys/param.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <vga.h>

#include <dev/videocode/videocode.h>
#include <dev/videocode/ioctl_video.h>

int modes16bpp[] = { G640x480x64K, G800x600x64K, G1024x768x64K};
int modes24bpp[] = { G640x480x16M, G800x600x16M, G1024x768x16M};

int
main(int argc, char *argv[])
{
	video_mode_t vm;
	vga_modeinfo *mi;
	FILE *f;
	int zrfd;
	int mode;
	int i;
	int g;
	int lsize;
	unsigned char *mem;
	unsigned long inp;
	struct decoder_buffer db;
	struct overlay_mask om;
	struct tvnorm tn;
	int ch;
	char *devname;
	int capture;
	int verbose;
	int norms;
	int rev;
	int overlay;
	void *obuf;
	int obuflen;

	/* Defaults */
	devname = "/dev/zr0";
	inp = 0;
	vm = VIDEO_MODE_AUTO;
	verbose = 0;
	capture = 1;
	obuf = NULL;

	while ((ch = getopt(argc, argv, "cf:i:m:nov")) != -1)
		switch (ch) {
		case 'c':
			capture = 0;
			break;
		case 'f':
			devname = optarg;
			break;
		case 'i':
			inp = atoi(optarg);
			break;
		case 'm':
			vm = atoi(optarg);
			break;
		case 'n':
			norms = 1;
			break;
		case 'o':
			overlay = 1;
			break;
		case '?':
		default:
			fprintf(stderr,
"Usage: ctv [ -f device ] [ -i inp ] [ -m mode ] [ -n ] [ -v ]\n\n"
"mode: 0 = AUTO, 1 = NTSC, 2 = PAL, 3 = SECAM\n"
"n: print out resolution information\n"
"v: verbose output\n");
			exit(1);
		}

	if ((zrfd = open(devname, O_RDWR, 0)) == -1)
		err(2, "%s", devname);

	if (ioctl(zrfd, DECODER_SET_INPUT, &inp) == -1)
		err(3, "DECODER_SET_INPUT");

	if (ioctl(zrfd, DECODER_SET_NORM, &vm) == -1)
		err(4, "DECODER_SET_NORM");
	if (verbose) {
		if (ioctl(zrfd, DECODER_GET_NORM, &vm) == -1)
			err(5, "DECODER_GET_NORM");
		printf("video mode: %s\n", video_mode_to_str(vm));
	}

	if (norms) {
		if (ioctl(zrfd, DECODER_GET_TVNORM, &tn) == -1) {
			perror("DECODER_GET_TVNORM");
		}

		printf("Width: %d, Height: %d\n", tn.Wa, tn.Ha);
	}

	if (verbose)
		f = stderr;

	vga_init();

	rev = 0;
again:
	g = 0;
	for (i = 0; i < sizeof modes16bpp / sizeof *modes16bpp; i++) {
		mode = modes16bpp[i];
		mi = vga_getmodeinfo(mode);

		if (mi == NULL || vga_hasmode(mode) != 0) {
			if ((mi->flags & CAPABLE_LINEAR) == CAPABLE_LINEAR ||
			    (mi->flags & IS_LINEAR) == IS_LINEAR)
				g = 1;

			if (verbose || (mi->flags & EXT_INFO_AVAILABLE)
			    == EXT_INFO_AVAILABLE)
				fprintf(f, "lin app: %p\n",
				    mi->linear_aperture);
		}
		if (!rev && (mi->width < tn.Wa || mi->height < tn.Ha)) {
			g = 0;
			rev = 1;
		}

		if (g)
			break;
	}

	if (i == sizeof modes16bpp / sizeof *modes16bpp) {
		if (rev)
			goto again;

		fprintf(f, "no modes!\n");
		exit(13);
	}

	if (capture) {
		if (vga_setmode(mode) != 0) {
			fprintf(f, "failed to set mode\n");
			exit(14);
		}

		if ((lsize = vga_setlinearaddressing()) == -1)
			errx(6, "vga_setlinearaddressing()\n");

		if (verbose)
			fprintf(f, "linear size: %d\n", lsize);

		mem = vga_getgraphmem();

		/* clear the display */
		memset(mem, 0, lsize);

		db.addr = mem;
		db.len = lsize;
		db.format = DECODER_GEO_RGB16;
		db.vidwidth = mi->linewidth;
		if (verbose) {
			fprintf(f, "linewidth: %d, bpp: %d\n", (int)mi->linewidth,
			    (int)mi->bytesperpixel);
			fprintf(f, "vidwidth: %d\n", db.vidwidth);
		}
		db.width = MIN(mi->width, tn.Wa);
		db.height = MIN(mi->height, tn.Ha);
		if (1) {
			db.xstart = (MAX(mi->width, tn.Wa) - db.width) / 2;
			db.ystart = (MAX(mi->height, tn.Ha) - db.height) / 2;
		} else
			db.ystart = db.xstart = 0;

		if (ioctl(zrfd, DECODER_SVIDBUF, &db) == -1)
			err(7, "DECODER_SVIDBUF");
	} else {
		struct decoder_geomet dg;

		dg.columns = MIN(mi->width, tn.Wa);
		dg.rows = MIN(mi->height, tn.Ha);
		dg.oformat = DECODER_GEO_RGB16;
		if (ioctl(zrfd, DECODER_SETGEO, &dg) == -1)
			err(8, "DECODER_SETGEO");
	}

	if (overlay) {
		memset(&om, 0, sizeof om);
		om.om_allvis = 1;
		om.om_hide = 1;
		om.om_cnt = 1;
		om.om_boxes[0] = (struct overlay_box) { 100, 100, 300, 300 };
		if (ioctl(zrfd, VIDEO_OVERLAY, &om) == -1) {
			vga_setmode(TEXT);
			printf("errno: %d\n", errno);
			perror("VIDEO_OVERLAY");
			sleep(2);
			err(9, "VIDEO_OVERLAY");
		}
	}

	if (overlay) {
		obuflen = (db.width * db.height + 7) / 8;
		obuf = mmap(NULL, obuflen, PROT_READ|PROT_WRITE, MAP_SHARED,
		    zrfd, 0);
		if (obuf == MAP_FAILED) {
			perror("mmap");
			sleep(2);
		}
	}

	vga_lockvc();

	if (capture) {
		i = DECODER_CAP_CONTINOUS;
		if (ioctl(zrfd, DECODER_CAPTURE, &i) == -1)
			err(10, "DECODER_CAPTURE");

		vga_getch();

		i = DECODER_CAP_STOP_CONT;
		if (ioctl(zrfd, DECODER_CAPTURE, &i) == -1)
			err(11, "DECODER_CAPTURE");
	}

	vga_unlockvc();

	if (capture)
		vga_setmode(TEXT);


	return 0;
}
