/*-
 * Copyright 2003 John-Mark Gurney
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: adv717xa.c,v 1.2 2003/04/30 16:31:48 jmg Exp jmg $
 */

/*
 * Cribbed from saa711x and iicsmb
 */

#include <sys/param.h>
#include <sys/kernel.h>
#include <sys/systm.h>
#include <sys/module.h>
#include <sys/uio.h>

#include <sys/bus.h>
#include "device_if.h"

#include <dev/iicbus/iiconf.h>
#include <dev/iicbus/iicbus.h>
#include "iicbus_if.h"

#include <dev/videocode/ioctl_video.h>
#include <dev/videocode/videocode.h>

#define IICBUS_TIMEOUT 100     /* us */

#define ADV_SOFTC						\
	struct adv717xa_softc *adv				\
	    = (struct adv717xa_softc *)device_get_softc(dev)

#define ADV7175A_SA_A	0xd4	/* if pin 18 (ALSB) is low */
#define ADV7175A_SA_B	0xd6	/* if pin 18 (ALSB) is high */
#define ADV7176A_SA_A	0x54	/* if pin 18 (ALSB) is low */
#define ADV7176A_SA_B	0x56	/* if pin 18 (ALSB) is high */

#define TR0_REG	0x07
#define TR0MODE	0x46
#define TR0RST	0x80

struct adv717xa_ivars {
	const char *desc;
	u_char	addr;
	int	maxinp;
	int	maxout;
};

struct adv717xa_softc {
	const struct adv717xa_ivars *adv_ivars;
#define adv_slvaddr	adv_ivars->addr
#define adv_maxinp	adv_ivars->maxinp
#define adv_maxout	adv_ivars->maxout

	u_int	adv_inp;
	video_mode_t	adv_mode;
	u_int	adv_outen	: 1;
	u_int	adv_rev		: 1;
	int	adv_bright;
	int	adv_contrast;
	int	adv_hue;
	int	adv_sat;

	device_t adv_iicbus;
	device_t adv_self;
};

#define DGET_BUS(dev)	iicbus_request_bus(dev, dev, IIC_WAIT)
#define GET_BUS(adv)	iicbus_request_bus(adv->adv_iicbus,	\
			    adv->adv_self, IIC_WAIT)
#define DREL_BUS(dev)	iicbus_release_bus(dev, dev)
#define REL_BUS(adv)	iicbus_release_bus(adv->adv_iicbus,	\
			    adv->adv_self)
#define RD_START(adv)	iicbus_start(adv->adv_iicbus,		\
			    adv->adv_slvaddr|1, IICBUS_TIMEOUT)
#define WR_START(adv)	iicbus_start(adv->adv_iicbus,		\
			    adv->adv_slvaddr, IICBUS_TIMEOUT)
#define STOP(adv)	iicbus_stop(adv->adv_iicbus)
#define DSTOP(dev)	iicbus_stop(dev)

static void adv717xa_disable_output(device_t dev);

static int
adv_write(struct adv717xa_softc *adv, u_char reg, u_char byte)
{
	int error = 0;

	if ((error = GET_BUS(adv)))
		return error;
	if ((error = WR_START(adv)))
		goto fail1;
	if ((error = iicbus_write_byte(adv->adv_iicbus, reg, IICBUS_TIMEOUT)))
		goto fail2;
	error = iicbus_write_byte(adv->adv_iicbus, byte, IICBUS_TIMEOUT);
fail2:
	STOP(adv);
fail1:
	REL_BUS(adv);
	return error;
}

static int
adv_read(struct adv717xa_softc *adv, u_char reg, u_char *byte)
{
	int error = 0;

	if ((error = GET_BUS(adv)))
		return error;
	if ((error = WR_START(adv)))
		goto fail1;
	if ((error = iicbus_write_byte(adv->adv_iicbus, reg, IICBUS_TIMEOUT)))
		goto fail2;
	if ((error = iicbus_repeated_start(adv->adv_iicbus, adv->adv_slvaddr | 1, IICBUS_TIMEOUT)))
		goto fail2;
	error = iicbus_read_byte(adv->adv_iicbus, byte, IICBUS_TIMEOUT);
fail2:
	STOP(adv);
fail1:
	REL_BUS(adv);
	if (error)
		device_printf(adv->adv_self, "read error: %d\n", error);
	return error;
}

static void
adv_rsttr0(device_t dev)
{
	ADV_SOFTC;

	adv_write(adv, TR0_REG, TR0MODE);
	adv_write(adv, TR0_REG, TR0MODE | TR0RST);
	adv_write(adv, TR0_REG, TR0MODE);
}

static const char adv7175a_desc[] = "Analog Devices ADV7175A (w/ Macrovision) Video Encoder";
static const char adv7176a_desc[] = "Analog Devices ADV7176A Video Encoder";
static const struct adv717xa_ivars chips[] = {
    { adv7175a_desc, ADV7175A_SA_A, 3, 3},
    { adv7175a_desc, ADV7175A_SA_B, 3, 3},
    { adv7176a_desc, ADV7176A_SA_A, 3, 3},
    { adv7176a_desc, ADV7176A_SA_B, 3, 3}
};

#define ADV_WRITE_BLOCK(error, adv, blk, all)				\
	do {								\
		int j;							\
		(error) = iicbus_block_write((adv)->adv_iicbus,		\
		/* Avoid a -Wcast-qual warning */			\
		    (adv)->adv_slvaddr, (u_char *)(uintptr_t)blk,	\
		    sizeof blk / sizeof *blk, &j);			\
		(all) = j == sizeof blk / sizeof *blk;			\
	} while (0)

static const u_char init_com1[] = { 0x00,
	0x11, 0x00, 0x0c, 0x8c, 0x79, 0x26, 0x40, 0x46,
	0x21, 0x00, 0x00, 0x00, 0x80, 0x4f, 0x00, 0x00,
	0x00, 0x00, 0x00 };
static const u_char init_com2[] = { 0x24, 0x00 };
static const u_char init_pal[] = { 0x00,
	0x11, 0x00, 0x0c, 0x8c, 0x79, 0x26, 0x40 };
static const u_char init_ntsc[] = { 0x00,
	0x14, 0x00, 0x55, 0x55, 0x55, 0x25, 0x1a };

static const struct {
	int	cnt;
	const u_char	*data;
} init_com[] = { { sizeof init_com1 / sizeof *init_com1, init_com1 },
	{ sizeof init_com2 / sizeof *init_com2, init_com2 },
};

static int
adv717xa_init_com(device_t dev)
{
	ADV_SOFTC;
	int i, j;
	int error;
#if 0
	u_char byteo, byten, bytel;

	adv_read(adv, 0x00, &byteo);
	adv_write(adv, 0x00, 0x11);
	adv_read(adv, 0x00, &byten);
	adv_write(adv, 0x00, 0x14);
	adv_read(adv, 0x00, &bytel);
	device_printf(dev, "byteo: %02x, byten: %02x, bytel: %02x\n", byteo,
	    byten, bytel);
#endif

	for (i = 0; i < sizeof init_com / sizeof *init_com; i++) {
		/* Avoid a -Wcast-qual warning */
		error = iicbus_block_write(adv->adv_iicbus, adv->adv_slvaddr,
		    (u_char *)(uintptr_t)init_com[i].data, init_com[i].cnt, &j);
		if (j != init_com[i].cnt || error) {
			device_printf(dev, "failed to init_com: %d\n", error);
			return error;
		}
	}

	adv_rsttr0(dev);

	return 0;
}

static void
adv717xa_identify(driver_t *driver, device_t parent)
{
	const struct adv717xa_ivars *adi;
	device_t dev;
	int error;
	int i, len;

	if ((error = DGET_BUS(parent)))
		return;

	DSTOP(parent);
	IICBUS_RESET(parent, IIC_FASTEST, 0, NULL);

	/* Probe for chips */
	len = sizeof chips / sizeof *chips;
	for (i = 0; i < len; i++) {
		if ((error = iicbus_start(parent, chips[i].addr,
		    IICBUS_TIMEOUT)))
			continue;

		DSTOP(parent);
		/* we have a chip that responded */
		adi = &chips[i];
		dev = device_add_child(parent, driver->name, -1);
		device_set_ivars(dev, (void *)(uintptr_t)adi);
	}

	IICBUS_RESET(parent, IIC_FASTEST, 0, NULL);
	DREL_BUS(parent);
}

static int
adv717xa_probe(device_t dev)
{
	struct adv717xa_ivars *adi = device_get_ivars(dev);
	device_t iicbus = device_get_parent(dev);
	ADV_SOFTC;
	u_char val;
#if 0
	device_t *sibs;
	driver_t *drv;
	int sibcnt;
	int error;
	int i, j, len;

	sibs = NULL;
	drv = device_get_driver(dev);
#endif

	if (adi == NULL)
		return ENXIO;

	memset(adv, 0, sizeof *adv);
	adv->adv_ivars = adi;
	adv->adv_iicbus = iicbus;
	adv->adv_self = dev;

#if 0
	if ((error = GET_BUS(adv)))
		return error;

	STOP(adv);
	IICBUS_RESET(iicbus, IIC_FASTEST, 0, NULL);

	/* Probe for chips */
	len = sizeof chips / sizeof *chips;
	for (i = 0; i < len; i++) {
		if ((error = iicbus_start(iicbus, chips[i].addr,
		    IICBUS_TIMEOUT))) {
			device_printf(dev, "failed to start bus\n");
			continue;
		}
		/* Make sure this isn't already attached */
		if (sibs == NULL)
			device_get_children(iicbus, &sibs, &sibcnt);
		for (j = 0; j < sibcnt; j++) {
			if (device_get_state(sibs[j]) >= DS_ATTACHED
			    && device_get_driver(sibs[j]) == drv
			    && ((struct adv717xa_softc *)
			    device_get_softc(sibs[j]))->adv_slvaddr
			    == chips[i].addr) {
				device_printf(dev, "already attached at %s\n",
				    device_get_nameunit(sibs[j]));
				break;
			}
		}

		if (j < sibcnt)
			continue;
		break;
	}

	if (i == len) {
		error = REL_BUS(adv);
		return ENXIO;
	}

	STOP(adv);

	IICBUS_RESET(iicbus, IIC_FASTEST, 0, NULL);

	if ((error = REL_BUS(adv)))
		device_printf(dev, "error releasing iicbus in probe\n");

	adv->adv_slvaddr = chips[i].addr;
	device_set_desc(dev, chips[i].desc);

	adv->adv_maxinp = chips[i].maxinp;
	adv->adv_maxout = chips[i].maxout;
#endif

	adv_read(adv, 0x12, &val);
	adv->adv_rev = val & 1;
	device_set_desc(dev, adi->desc);

	return 0;
}

static int
adv717xa_attach(device_t dev)
{
	ADV_SOFTC;
	int error;

	if ((error = adv717xa_init_com(dev)))
		return error;

	if (adv == NULL)
		device_printf(dev, "softc is NULL!!\n");
	else {
		device_printf(dev, "attach slave: %02x, rev: %d\n", (int)adv->adv_slvaddr, (int)adv->adv_rev);
	}

	return 0;
}

static int
adv717xa_detach(device_t dev)
{
	adv717xa_disable_output(dev);
	device_set_ivars(dev, NULL);
	return 0;
}

static void
adv717xa_get_cap(device_t dev, int *flags, int *inps, int *outs)
{
	ADV_SOFTC;

	*flags = VIDEO_PAL | VIDEO_NTSC;
	*inps = adv->adv_maxinp;
	*outs = adv->adv_maxout;
}

static video_mode_t
adv717xa_set_mode(device_t dev, video_mode_t mode)
{
	ADV_SOFTC;
	int error;
	int all;

	/*
	 * table w/ data so I can just loop over the table?
	 */
	switch (mode) {
	case VIDEO_MODE_NTSC:
		ADV_WRITE_BLOCK(error, adv, init_ntsc, all);
		break;
	case VIDEO_MODE_PAL:
		ADV_WRITE_BLOCK(error, adv, init_pal, all);
		break;
	default:
		return VIDEO_MODE_ERROR;
	}

	if (error || !all)
		return VIDEO_MODE_ERROR;

	adv_rsttr0(dev);

	adv->adv_mode = mode;
	return mode;
}

static void
adv717xa_set_inp(device_t dev, int inp)
{
	ADV_SOFTC;

	switch (inp) {
	case 0:	/* from decoder */
		adv_write(adv, 0x01, 0x00);
		adv_write(adv, 0x0c, 0x80);
		adv_write(adv, 0x0d, 0x4f);
		break;
	case 1:	/* from controller */
		adv_write(adv, 0x01, 0x00);
		adv_write(adv, 0x0c, 0x00);
		adv_write(adv, 0x0d, 0x49);
		break;
	case 2:	/* color bars */
		adv_write(adv, 0x01, 0x80);
		adv_write(adv, 0x0d, 0x49);
		break;
	}
	adv_rsttr0(dev);
}

static void
adv717xa_enable_output(device_t dev)
{
	ADV_SOFTC;
	u_char val;

	if (adv_read(adv, 0x01, &val))
		return;
	val |= 0xf << 3;
	if (adv_write(adv, 0x01, val))
		return;

	adv->adv_outen = 0;
}

static void
adv717xa_disable_output(device_t dev)
{
	ADV_SOFTC;
	u_char val;

	if (adv_read(adv, 0x01, &val))
		return;
	val &= ~(0xf << 3);
	if (adv_write(adv, 0x01, val))
		return;

	adv->adv_outen = 0;
}

static devclass_t adv717xa_devclass;

static device_method_t adv717xa_methods[] = {
	/* device interface */
	DEVMETHOD(device_identify,	adv717xa_identify),
	DEVMETHOD(device_probe,		adv717xa_probe),
	DEVMETHOD(device_attach,	adv717xa_attach),
	DEVMETHOD(device_detach,	adv717xa_detach),

	/* videocode interface */
	DEVMETHOD(videocode_get_cap,		adv717xa_get_cap),
	DEVMETHOD(videocode_set_mode,		adv717xa_set_mode),
	DEVMETHOD(videocode_set_inp,		adv717xa_set_inp),
	/*DEVMETHOD(videocode_set_picture,	adv717xa_set_picture),*/
	DEVMETHOD(videocode_enable_output,	adv717xa_enable_output),
	DEVMETHOD(videocode_disable_output,	adv717xa_disable_output),
	DEVMETHOD(videocode_istype,		videocode_isencoder),

#if 0
	/* I don't think I need this */
	/* iicbus interface */
        DEVMETHOD(iicbus_intr,          iicsmb_intr),
#endif

	{ 0, 0}
};

static driver_t adv717xa_driver = {
        "adv717xa",
        adv717xa_methods,
	sizeof(struct adv717xa_softc)
};

DRIVER_MODULE(adv717xa, iicbus, adv717xa_driver, adv717xa_devclass, 0, 0);
MODULE_DEPEND(adv717xa, iicbus, IICBUS_MINVER, IICBUS_PREFVER, IICBUS_MAXVER);
MODULE_DEPEND(adv717xa, videocode, VIDEOCODE_MINVER, VIDEOCODE_PREFVER,
	VIDEOCODE_MAXVER);
MODULE_VERSION(adv717xa, 1);
