/*-
 * Copyright 2003 John-Mark Gurney
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: videocode.c,v 1.1 2003/04/30 16:28:48 jmg Exp jmg $
 */

#include <sys/param.h>
#include <sys/kernel.h>
#include <sys/systm.h>
#include <sys/module.h>
#include <sys/malloc.h>
#include <sys/bus.h>

#include <dev/videocode/ioctl_video.h>
#include <dev/videocode/videocode.h>

static moduledata_t mod_data = {
	"videocode",
	NULL,	/* we use the DEPEND to do our ref counting for us. */
	0
};

static const char *mode_strs[] = { "error", "NTSC", "PAL", "SECAM", "auto" };

const char *
video_mode_str(video_mode_t mode)
{
	if (mode < 0 || mode > sizeof mode_strs / sizeof *mode_strs)
		return "invalid mode";

	return mode_strs[mode];
}

static int
videocode_findtype_internal(device_t dev, int type, device_t **devs, int *cnt)
{
	device_t *devlist;
	int i, c, error;

	if ((error = VIDEOCODE_ISTYPE(dev)) == type) {
		/* found one! */
		/*printf("vc: found: dev: %p, type: %d\n", dev, type);*/
		*devs = realloc(*devs, *cnt * sizeof **devs, M_TEMP, M_WAITOK);
		*devs[(*cnt)++] = dev;
	}

	if ((error = device_get_children(dev, &devlist, &c)))
		return error;

	error = 0;
	for (i = 0; error == 0 && i < c; i++)
		error = videocode_findtype_internal(devlist[i], type, devs,
		    cnt);

	free(devlist, M_TEMP);

	return error;
}

/*
 * Searches all children of dev for videocode device of type
 * the devs array will be allocated for you and must be freed w/
 * free(ar, M_TEMP)
 * cnt will return total number of type of devices
 * even in an error if devs is not NULL, it must be freed as it will/may
 * contain a partial list.
 */
int
videocode_findtype_alloc(device_t dev, int type, device_t **devs, int *cnt)
{
	if (cnt == NULL || devs == NULL)
		return EFAULT;

	*devs = NULL;
	*cnt = 0;

	return videocode_findtype_internal(dev, type, devs, cnt);
}

int
videocode_findtype(device_t dev, int type, device_t *devs, int *cnt)
{
	device_t *dp;
	int c, error, i;

	error = videocode_findtype_alloc(dev, type, &dp, &c);

	if (error)
		printf("vc: findtype_alloc error: %d\n", error);

	if (c) {
		for (i = 0; i < c && i < *cnt; i++)
			devs[i] = dp[i];
		if (*cnt == i)
			/*
			 * we got enough to satisfy the request,
			 * hide any errors
			 */
			error = 0;
		*cnt = i;
	}

	free(dp, M_TEMP);

	return error;
}

int
videocode_isencoder(device_t dev)
{
	return VIDEO_ENCODER;
}

int
videocode_isdecoder(device_t dev)
{
	return VIDEO_DECODER;
}

DECLARE_MODULE(videocode, mod_data, SI_SUB_EXEC, SI_ORDER_ANY);
MODULE_VERSION(videocode, VIDEOCODE_MAXVER);
