/*-
 * Copyright 2003 John-Mark Gurney
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: zr_core.c,v 1.21.2.1 2003/05/08 03:36:14 jmg Exp $
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/lock.h>
#include <sys/mutex.h>
#include <sys/malloc.h>
#include <sys/ioccom.h>
#include <sys/bus.h>  /* structs, prototypes for pci bus stuff */
#include <machine/bus_memio.h>
#include <machine/bus.h>
#include <sys/rman.h>

#include <dev/videocode/ioctl_video.h>
#include <dev/videocode/videocode.h>

#include <dev/postoffice/postoffice.h>

#include "opt_zr.h"
#include <dev/zr/zr_reg.h>
#include <dev/zr/zr_core.h>
#include <dev/zr/zr_os.h>

#if ZR_METEOR_COMPAT
#include <machine/ioctl_meteor.h>
#endif

/* f = 1111
   e = 1110	7 = 0111
   c = 1100	3 = 0011
   8 = 1000	1 = 0001
*/
static u_char head_mask[8] = { 0xff, 0xfe, 0xfc, 0xf8, 0xf0, 0xe0, 0xc0, 0x08 };
static u_char tail_mask[8] = { 0x01, 0x03, 0x07, 0x0f, 0x1f, 0x3f, 0x7f, 0xff };

#define ZR_DUMB_GOTO(zr, st)						\
	do {								\
		int _error;						\
		if ((_error = zr_goto_state((zr), (st))))		\
			ZRSC_PRINTF((zr),				\
	"uncaught state change error! from: %d, to %d, error: %d\n",	\
			    (zr)->zr_state, (st), _error);		\
	} while (0)

const struct vendor_info vendors[] = {
#define PINMICRO	&vendors[0]
	{ "Pinnacle Systems", 0x1031 },
};
const char svideo[] = "S-Video";
const char composite[] = "Composite";
const char *pinmicroputs[] = { composite, svideo, "Composite/Internal" };
const int dc10inps[] = { 0, 5, 7 };

const struct zr_carddesc cards[] = {
	{ .desc = "DC10+",
	  .vi = PINMICRO,
	  .sysid = 0x7efe,
	  .revid = 2,
	  .irq_vsync = 1,
	  .gpio_dir = 0x00,
	  .gpio_setframe = 6,
	  .gpio_setvidbusdir = 1,
	  .gpio_setvidbusen = 2,
	  .gpio_freq = 5,
	  .HSPol = 1,	
	  .VSPol = 1,
	  .indesc = pinmicroputs,
	  .outdesc = pinmicroputs,
	  .inchannels = 3,
	  .outchannels = 3,
	  .inchans = dc10inps,
	  .outchans = dc10inps, /* XXX - when output is supported fix chans */
	  .subchips = {
		{ .desc = "ZR36060",
		  .subid = 0,
		  .irq = 0,
		  .reset = 0,
		  .sleep = 3 },
		{ .desc = NULL } }
	}
};

#define ZR_DUMPREGS(zr, regs)	zr_dumpregs(zr, (regs),			\
					sizeof(regs) / sizeof *(regs))
static const int vfe_regs[] = {
	ZR_VFEHCR_REG,
	ZR_VFEVCR_REG,
	ZR_VFESPFR_REG,
	ZR_VDTR_REG,
	ZR_VDBR_REG,
	ZR_VSSFGR_REG,
	ZR_VDCR_REG,
	ZR_MMTR_REG,
	ZR_MMBR_REG,
	ZR_OCR_REG
};

static const int control_regs[] = {
	ZR_SYSPCIGPIO_REG,
	ZR_GPPR_REG,
	ZR_ISR_REG,
	ZR_ICR_REG,
	ZR_GBCR2_REG
};

static int
printputs(const char **list, int cnt)
{
	int i;
	int ret;

	for (ret = i = 0; i < cnt; i++)
		ret += printf("%s%s", i ? ", " : "", list[i]);

	return ret;
}

static void
zr_dumpregs(struct zr_softc *zr, const int *regs, int cnt)
{
#define NCOLS	4
	int i;
	int j;
	int r[NCOLS];
	int end;
	u_int32_t b[NCOLS];

	for (i = 0; i < cnt; i++) {
		b[i % NCOLS] = zr_read_asrs(zr, (r[i % NCOLS] = regs[i]));
		if (i % NCOLS == (NCOLS - 1) || i + 1 == cnt) {
			if (cnt - ((i / NCOLS) * NCOLS) < NCOLS)
				end = cnt - ((i / NCOLS) * NCOLS);
			else
				end = NCOLS;
			for (j = 0; j < end; j++)
				printf("0x%03x: %08x%s", r[j], b[j],
				    (j == (end - 1)) ? "\n" : ", ");
		}
	}
#undef NCOLS
}

int
zr_goto_state(struct zr_softc *zr, zr_state_t st)
{
	int error;

	error = 0;

	mtx_lock(&zr->zr_lock);

	if (st == zr->zr_state)
		error = EBUSY;
	else switch (st) {
	case ZR_DISABLED:
		if (zr->zr_opened || !zr->zr_enabled)
			error = EACCES;
		else {
			zr->zr_enabled = 0;
			zr->zr_state = ZR_DISABLED;
		}
		break;

	case ZR_ENABLED:
		if (zr->zr_enabled)
			error = EACCES;
		else {
			zr->zr_enabled = 1;
			zr->zr_state = ZR_ENABLED;
		}
		break;

	case ZR_DEVOPENED:
		if (zr->zr_opened)
			error = EBUSY;
		else if (!zr->zr_enabled)
			error = EACCES;
		else {
			zr->zr_opened = 1;
			zr->zr_state = ZR_DEVOPENED;
			zr->zr_bufset = 0;
			zr->zr_winset = 0;
		}
		break;

	case ZR_DEVCLOSED:
		if (!zr->zr_opened)
			error = EACCES;
		else {
			zr->zr_opened = 0;
			zr->zr_state = ZR_ENABLED;
		}
		break;

	case ZR_CAPBUF:
		if (!zr->zr_opened || zr->zr_nbufs != 0)
			error = EACCES;
		else {
			zr->zr_capbuf = 1;
			if (zr->zr_nbufs == 4)
				error = EBUSY; /* ENOSPC? */
			else
				zr->zr_nbufs++;
		}
		break;

	case ZR_CAPBUFFREE:
		if (!zr->zr_opened || zr->zr_nbufs == 0 || !zr->zr_capbuf)
			error = EACCES;
		else
			zr->zr_nbufs--;
		break;

	case ZR_CAPTURE_START:
		/*ZRSC_PRINTF(zr, "opened: %d, bufset: %d, winset: %d\n",
		    zr->zr_opened, zr->zr_bufset, zr->zr_winset);*/
		if (!zr->zr_opened || !zr->zr_bufset || !zr->zr_winset)
			error = EACCES;
		else if (zr->zr_capture)
			error = EBUSY;
		else
			zr->zr_capture = 1;
		break;

	case ZR_CAPTURE_STOP:
		/*ZRSC_PRINTF(zr, "capture: %d\n", zr->zr_capture);*/
		if (!zr->zr_capture)
			error = EACCES;
		else
			zr->zr_capture = 0;
		break;

	case ZR_BUFSET:
		if (!zr->zr_opened || zr->zr_capture)
			error = EACCES;
		else
			zr->zr_bufset = 1;
		break;

	case ZR_WINSET:
		if (!zr->zr_opened || zr->zr_capture)
			error = EACCES;
		else
			zr->zr_winset = 1;
		break;

	case ZR_MASKALLOC:
		if (!zr->zr_opened || zr->zr_capture || zr->zr_maskalloc ||
		    zr->zr_mask)
			error = EACCES;
		else
			zr->zr_maskalloc = 1;
		break;

	case ZR_MASK:
		if (!zr->zr_maskalloc)
			error = EACCES;
		else
			zr->zr_mask = 1;
		break;

	case ZR_MASKFREE:
		if (!zr->zr_maskalloc || zr->zr_mask)
			error = EACCES;
		else
			zr->zr_maskalloc = 0;
		break;

	case ZR_MASKDEALLOC:
		if (!zr->zr_mask || !zr->zr_maskalloc || zr->zr_capture)
			error = EACCES;
		else
			zr->zr_mask = 0;
		break;
	}

	mtx_unlock(&zr->zr_lock);

	return error;
}

static void
zr_after_reset(struct zr_softc *zr)
{
	static const int regs[] = { ZR_GPPR_REG, ZR_GBCR2_REG };
	int reg;
	int i;
	int wr;
	asrs_t asr;

	/*ZRSC_PRINTF(zr, "after_reset\n");*/

	reg = wr = -1;

	/*
	 * make it an output since we can only set this GenPurDir once after
	 * reset.
	 */
	if (zr->zr_card != NULL) {
		zr_write_asrs(zr, ZR_SYSPCIGPIO_REG, (zr_read_asrs(zr,
		    ZR_SYSPCIGPIO_REG) & ~(0xff)) | zr->zr_card->gpio_dir);
		/*ZRSC_PRINTF(zr, "SYSGPIO: 0x%08x\n", zr_read_asrs(zr,
		    ZR_SYSPCIGPIO_REG));*/
	}

	/*
	 * setup the timing parameters for the GuestBus
	 */
	for (i = 0; i < 8; i++) {
		if (i == 0 || i == 4) {
			wr = 0;
			reg = regs[i / 4];
			asr.reg = zr_read_asrs(zr, reg) & ~(0xffff);
		}

		if (zr->zr_gb_dur[i] || zr->zr_gb_rec[i]) {
			/* we have something to add */
			asr.reg |= ((zr->zr_gb_dur[i] & 0x3) << 2
			    | (zr->zr_gb_rec[i] & 0x3)) << ((i % 4) * 4);
			wr = 1;
		}

		if ((i == 3 || i == 7)) {
			if (wr)
				zr_write_asrs(zr, reg, asr.reg);
			/*ZRSC_PRINTF(zr, "reg: 0x%03x, val: 0x%x\n", reg,
			    asr.reg);*/
		}
	}

	ZR_SET_ASR(zr, ZR_GPPR_REG, gppr, GenPurIO, 0x81);

	if (zr->zr_card != NULL) {
		zr_gpio(zr, zr->zr_card->gpio_setvidbusdir, 0);
		zr_gpio(zr, zr->zr_card->gpio_setvidbusen, 0);
		zr_gpio(zr, zr->zr_card->gpio_setframe, 1);
	}
}

void
zr_reset(struct zr_softc *zr)
{
	/*ZRSC_PRINTF(zr, "reset\n");*/

	zr_disable(zr);
	DELAY(1);		/* the sheet doesn't say a delay is needed */
	zr_enable(zr);		/* after_reset gets called by enable */
}

void
zr_enable(struct zr_softc *zr)
{
	int error;
	uint32_t v;

	if ((error = zr_goto_state(zr, ZR_ENABLED))) {
		ZRSC_PRINTF(zr, "goto enabled error: %d\n", error);
		return;
	}

	v = zr_read_asrs(zr, ZR_SYSPCIGPIO_REG);
	v |= ZR_SOFTRESET;
	zr_write_asrs(zr, ZR_SYSPCIGPIO_REG, v);

	zr_after_reset(zr);
}

void
zr_disable(struct zr_softc *zr)
{
	zr_write_asrs(zr, ZR_SYSPCIGPIO_REG, 0);
	ZR_DUMB_GOTO(zr, ZR_DISABLED);
}
void
zr_enable_ints(struct zr_softc *zr)
{
	ZR_SET_ASR(zr, ZR_ICR_REG, icr, IntPinEn, 1);
}

void
zr_disable_ints(struct zr_softc *zr)
{
	ZR_SET_ASR(zr, ZR_ICR_REG, icr, IntPinEn, 0);
}

int
zr_mask_free(struct zr_softc *zr)
{
	int error;

	if ((error = zr_goto_state(zr, ZR_MASKDEALLOC)))
		goto fail;

	if ((error = zr_free_mem(zr, &zr->zr_bus.zb_mask))) {
		ZR_DUMB_GOTO(zr, ZR_MASK);
		goto fail;
	}

	ZR_DUMB_GOTO(zr, ZR_MASKFREE);

fail:
	return error;
}

/*
 * return 0 if success, EIO if failed
 */
static int
zr_po_wait(struct zr_softc *zr, asrs_t *asr)
{
	asrs_t rasr;
	if (asr == NULL)
		asr = &rasr;

	do {
		asr->reg = zr_read_asrs(zr, ZR_POR_REG);
		if (asr->por.POTime)
			break;
	} while (asr->por.POPen);

	return asr->por.POTime ? EIO : 0;
}

int
zr_po_read(struct zr_softc *zr, u_int8_t id, u_int8_t reg, u_int8_t *data)
{
	asrs_t asr;

	asr.reg = 0;
	asr.por = (struct por) {
	    .PODir =		0,
	    .POTime =		1,
	    .POGuestID =	id,
	    .POGuestReg =	reg
	};
	zr_write_asrs(zr, ZR_POR_REG, asr.reg);
	if (zr_po_wait(zr, &asr) != 0)
		return EIO;
	*data = asr.por.POData;

	return 0;
}

int
zr_po_write(struct zr_softc *zr, u_int8_t id, u_int8_t reg, u_int8_t data)
{
	asrs_t asr;

	asr.reg = 0;
	asr.por = (struct por) {
	    .PODir =		1,
	    .POTime =		1,
	    .POGuestID =	id,
	    .POGuestReg =	reg,
	    .POData =		data
	};
	zr_write_asrs(zr, ZR_POR_REG, asr.reg);
	return zr_po_wait(zr, NULL);
}

int
zr_core_attach(struct zr_softc *zr)
{
	int i;

	zr->zr_dec_mode = VIDEO_MODE_ERROR;

	for (i = 0; i < sizeof cards / sizeof *cards; i++) {
		if (zr->zr_subvenid == cards[i].vi->id
		    && zr->zr_subsysid == cards[i].sysid) {
			zr->zr_card = &cards[i];
			break;
		}
	}

	if (i == sizeof cards / sizeof *cards)
		return ENXIO;

	if (bootverbose) {
		ZRSC_PRINTF(zr, "Card: %s, Vendor: %s, In's: <",
		    zr->zr_card->desc, zr->zr_card->vi->desc);
		printputs(zr->zr_card->indesc, zr->zr_card->inchannels);
		printf(">, Out's: <");
		printputs(zr->zr_card->outdesc, zr->zr_card->outchannels);
		printf(">\n");
	}

	zr_reset(zr);

	return 0;
}

static int
get_bpp(u_int format)
{
	switch (format) {
	case DECODER_GEO_YUV_422:
	case DECODER_GEO_RGB16:
	case DECODER_GEO_RGB15:
		return 2;
	case DECODER_GEO_RGB24:
		return 3;
	case DECODER_GEO_RGB32:
		return 4;
	}

	return -1;
}

static int
get_YUV2RGB(u_int format)
{
	switch (format) {
	case DECODER_GEO_YUV_422:
		return 0;
	case DECODER_GEO_RGB32:
	case DECODER_GEO_RGB24:
		return 1;
	case DECODER_GEO_RGB16:
		return 2;
	case DECODER_GEO_RGB15:
		return 3;
	}

	return -1;
}

static int
zr_set_vfe(struct zr_softc *zr)
{
	asrs_t asr;
	struct decoder_buffer db;
	struct tvnorm tn;
	int bpp;
	int Wt, Wa, Ht, Ha;
	int HStart, HEnd, VStart, VEnd;
	int VidWinWid, VidWinHt;
	int hcrop1, hcrop2, vcrop1, vcrop2;
	int HorDcm, VerDcm;
	int X, We, Y, He;
	int DispMod;
	u_int format;

	if (zr->zr_dec_mode == VIDEO_MODE_ERROR)
		return EINVAL;

	/* XXX - TODO: make sure xstart and length are 4 byte aligned */
	db = zr->zr_buffer;
	format = db.format & DECODER_GEO_OUTPUT_MASK;
	bpp = get_bpp(format);
	if (db.xstart < 0 || db.ystart < 0 || db.width < 0 || db.vidwidth < 0
	    || db.height < 0)
		return EINVAL;
	if (bpp == -1 || (db.width + db.xstart) * bpp > db.vidwidth)
		return EINVAL;
	if ((db.xstart + db.width) * bpp
	    + (db.ystart + db.height) * db.vidwidth > db.len)
		return EINVAL;
	if (zr->zr_dec_norm == NULL)
		return ENXIO;

	tn = *zr->zr_dec_norm;

	if (tn.Wa < db.width || tn.Ha < db.height)
		return EINVAL;

	/* Givens */
	Wt = tn.Wt; Wa = tn.Wa;
	Ht = tn.Ht; Ha = tn.Ha;
	HStart = tn.HStart;
	VStart = tn.VStart;

	DispMod = !(db.height > Ha / 2);
	VidWinWid = db.width;
	VidWinHt = DispMod ? db.height : db.height / 2;

	/* First Eqs */
	HEnd = Wa + HStart - 1;
	VEnd = Ha / 2 + VStart - 1;

	/* Horizontal Eqs */
	X = (VidWinWid * 64 + Wa - 1) / Wa;	/* ceil(VidWinWid*64/Wa) */
	We = VidWinWid * 64 / X;		/* floor(VidWinWid*64/X) */
	HorDcm = 64 - X;
	hcrop1 = 2 * ((Wa - We) / 4);		/* 2*floor((Wa-We)/4) */
	hcrop2 = Wa - We - hcrop1;

	/*ZRSC_PRINTF(zr, "HStart': %d, HEnd': %d\n", HStart, HEnd);*/

	HStart = HStart + hcrop1;
	HEnd = HEnd - hcrop2;

	/*ZRSC_PRINTF(zr, "VidWinWid: %d, Wa: %d, HStart: %d, HEnd: %d, X: %d, We: %d, HorDcm: %d, hcrop1: %d, hcrop2: %d\n", VidWinWid, Wa, HStart, HEnd, X, We, HorDcm, hcrop1, hcrop2);*/

	/* Vertical Eqs */
	Y = ((VidWinHt * 64 * 2) + Ha - 1) / Ha; /* ceil(VidWinHt*64*2/Ha) */
	He = VidWinHt * 64 / Y;			/* floor(VidWinHt*64/Y) */
	VerDcm = 64 - Y;
	vcrop1 = (Ha / 2 - He) / 2;		/* floor((Ha/2-He)/2) */
	vcrop2 = Ha / 2 - He - vcrop1;

	/*ZRSC_PRINTF(zr, "VStart': %d, VEnd': %d\n", VStart, VEnd);*/

	VStart = VStart + vcrop1;
	VEnd = VEnd - vcrop2;

	/*ZRSC_PRINTF(zr, "VidWinHt: %d, Ha: %d, VStart: %d, VEnd: %d, X: %d, He: %d, VerDcm: %d, vcrop1: %d, vcrop2: %d\n", VidWinHt, Ha, VStart, VEnd, X, He, VerDcm, vcrop1, vcrop2);*/

	/* Set the registers */
	asr.reg = 0;
	asr.vfehcr.HSPol = zr->zr_card->HSPol;
	asr.vfehcr.HStart = HStart;
	asr.vfehcr.HEnd = HEnd;
	/*ZRSC_PRINTF(zr, "HSPol: %u, HStart: %u, HEnd: %u\n", asr.vfehcr.HSPol,
	    asr.vfehcr.HStart, asr.vfehcr.HEnd);*/
	zr_write_asrs(zr, ZR_VFEHCR_REG, asr.reg);

	asr.reg = 0;
	asr.vfevcr.VSPol = zr->zr_card->VSPol;
	asr.vfevcr.VStart = VStart;
	asr.vfevcr.VEnd = VEnd;
	/*ZRSC_PRINTF(zr, "VSPol: %u, VStart: %u, VEnd: %u\n", asr.vfevcr.VSPol,
	    asr.vfevcr.VStart, asr.vfevcr.VEnd);*/
	zr_write_asrs(zr, ZR_VFEVCR_REG, asr.reg);

	asr.reg = 0;
	if (zr->zr_dec_mode == VIDEO_MODE_NTSC)
		asr.vfespfr.ExtFl = 0;	
	else
		asr.vfespfr.ExtFl = 1;
	asr.vfespfr.TopField = 1;
	asr.vfespfr.VCLKPol = 0;
	asr.vfespfr.HFilter = HorDcm / 16;
	asr.vfespfr.DupFld = 0;
	asr.vfespfr.HorDcm = HorDcm;
	asr.vfespfr.VerDcm = VerDcm;
	asr.vfespfr.DispMod = DispMod;
	asr.vfespfr.YUV2RGB = get_YUV2RGB(format);
	asr.vfespfr.ErrDif = (format == DECODER_GEO_RGB15
	    || format == DECODER_GEO_RGB16) ? 1 : 0;
	asr.vfespfr.Pack24 = (format == DECODER_GEO_RGB32) ? 0 : 1;
	asr.vfespfr.LittleEndian = 1;
	/*ZRSC_PRINTF(zr,
	    "YUV2RGB: %u, DispMod: %u, VerDcm: %u, HorDcm: %u, HFilter: %u\n",
	    asr.vfespfr.YUV2RGB, asr.vfespfr.DispMod, asr.vfespfr.VerDcm,
	    asr.vfespfr.HorDcm, asr.vfespfr.HFilter);*/
	zr_write_asrs(zr, ZR_VFESPFR_REG, asr.reg);

	/* buffer settings */
	asr.reg = zr->zr_physaddr + (db.xstart * bpp
	    + db.ystart * db.vidwidth);
	/*ZRSC_PRINTF(zr, "VDTR: 0x%x\n", asr.reg);*/
	zr_write_asrs(zr, ZR_VDTR_REG, asr.reg);
	asr.reg += (DispMod ? 0 : db.vidwidth);
	/*ZRSC_PRINTF(zr, "VDBR: 0x%x\n", asr.reg);*/
	zr_write_asrs(zr, ZR_VDBR_REG, asr.reg);

	asr.reg = 0;
	asr.vssfgr.DispStride = (db.vidwidth - (db.width * bpp))
	    + (DispMod ? 0 : db.vidwidth);
	asr.vssfgr.VidOvf = 1;
	asr.vssfgr.SnapShot = 0;
	asr.vssfgr.FrameGrab = 0;
	/*ZRSC_PRINTF(zr, "DispMod: %d, DispStride: %u\n", DispMod,
	    asr.vssfgr.DispStride);
	ZRSC_PRINTF(zr,
	    "db.vidwidth: %d, db.xstart: %d, db.width: %d, bpp: %d\n",
	    db.vidwidth, db.xstart, db.width, bpp);*/
	zr_write_asrs(zr, ZR_VSSFGR_REG, asr.reg);

	asr.reg = zr_read_asrs(zr, ZR_VDCR_REG);
	asr.vdcr.MinPix = 8;	/* really twice due to Triton sharing */
	asr.vdcr.Triton = (zr->zr_revid <= 1) ? 0 : 1;
	asr.vdcr.VidWinHt = VidWinHt;
	asr.vdcr.VidWinWid = VidWinWid;
	/*ZRSC_PRINTF(zr, "VidWinHt: %u, VidWinWid: %u\n", asr.vdcr.VidWinHt,
	    asr.vdcr.VidWinWid);*/
	zr_write_asrs(zr, ZR_VDCR_REG, asr.reg);

	asr.reg = zr_read_asrs(zr, ZR_OCR_REG);
	asr.ocr.OvlEnable = zr->zr_mask && zr->zr_maskvalid ? 1 : 0;
	asr.ocr.MaskStride = DispMod == 0 ? zr->zr_maskstride / 4 : 0;
	zr_write_asrs(zr, ZR_OCR_REG, asr.reg);

	/*ZRSC_PRINTF(zr, "OvlEnable: %d, MaskStride: %d\n",
	    (int)asr.ocr.OvlEnable, (int)asr.ocr.MaskStride);*/

	if (asr.ocr.OvlEnable) {
		asr.reg = zr_dmamem_to_phys(zr, &zr->zr_bus.zb_mask);
		zr_write_asrs(zr, ZR_MMTR_REG, asr.reg);
		/*ZRSC_PRINTF(zr, "MMTR: 0x%x\n", asr.reg);*/

		asr.reg += zr->zr_maskstride;
		zr_write_asrs(zr, ZR_MMBR_REG, asr.reg);
		/*ZRSC_PRINTF(zr, "MMBR: 0x%x\n", asr.reg);*/
	}

	return 0;
}

int
zr_ioctl(struct zr_softc *zr, u_long cmd, caddr_t data, os_proc_t proc)
{
	int error;
	int i, j, k, ms;
	unsigned long inp;
	caddr_t mask;
	video_mode_t vm;
	struct decoder_buffer *db;
	u_int32_t phys;
	struct decoder_geomet *dg;
	struct decoder_vram *dv;
	struct tvnorm *tnp;
	struct overlay_mask *omp;
	struct overlay_box ob;

	error = 0;

	switch (cmd) {
	case DECODER_SET_INPUT:
		inp = *(unsigned long *)data;

		if (zr->zr_capture) {
			error = EBUSY;
			break;
		}

		IFCODERINVAL(zr, decoder) {
			error = ENXIO;
			break;
		}

		if (zr->zr_card == NULL) {
			error = ENXIO;
			break;
		}

#if ZR_METEOR_COMPAT
		if (inp == METEOR_INPUT_DEV0)
			inp = 0;
		else if (inp == METEOR_INPUT_DEV1)
			inp = 1;
		else if (inp == METEOR_INPUT_DEV2)
			inp = 2;
		else if (inp == METEOR_INPUT_DEV3)
			inp = 3;
		else if (inp == METEOR_INPUT_DEV_RGB)
			inp = 4;
		else if (inp == METEOR_INPUT_DEV_SVIDEO)
			for (i = 0; i < zr->zr_card->inchannels; i++)
				if (strcmp(zr->zr_card->indesc, svideo) == 0) {
					inp = i;
					break;
				}
#endif /* ZR_METEOR_COMPAT */

		if (inp >= zr->zr_card->inchannels) {
			error = EINVAL;
			break;
		}

		/*ZRSC_PRINTF(zr, "set input: %lu\n", inp);*/

		zr->zr_dec_inp = inp;
		error = VIDEOCODE_SET_INP(zr->zr_decoder,
		    zr->zr_card->inchans[inp]);

		break;

	case DECODER_GET_INPUT:
		*(unsigned long *)data = zr->zr_dec_inp;

		break;

	case DECODER_SET_NORM:
		vm = *(video_mode_t *)data;

		if (zr->zr_capture) {
			error = EBUSY;
			break;
		}

		if (vm < VIDEO_MODE_MIN || vm > VIDEO_MODE_MAX) {
			error = EINVAL;
			break;
		}

		IFCODERINVAL(zr, decoder) {
			error = ENXIO;
			break;
		}

		vm = VIDEOCODE_SET_MODE(zr->zr_decoder, vm);
		/*ZRSC_PRINTF(zr, "decoder mode: %s\n", video_mode_str(vm));*/

		if (vm == VIDEO_MODE_ERROR) {
			error = EINVAL;
			break;
		}

		zr->zr_dec_mode = vm;
		zr_set_norm(zr, VIDEOCODE_GET_TVNORM(zr->zr_decoder), 0);
		break;

	case DECODER_GET_NORM:
		if (zr->zr_dec_mode == VIDEO_MODE_ERROR)
			error = ENXIO;
		else
			*(enum video_mode *)data = zr->zr_dec_mode;
		break;

	case DECODER_SET_TVNORM:
		if (zr->zr_dec_mode == VIDEO_MODE_ERROR) {
			error = ENXIO;
			break;
		}
		tnp = (struct tvnorm *)data;

		if (tnp->Wt != zr->zr_dec_norm->Wt ||
		    tnp->Ht != zr->zr_dec_norm->Ht ||
		    tnp->HSyncStart != zr->zr_dec_norm->HSyncStart) {
			error = EINVAL;
			break;
		}

		zr->zr_winset = 0;
		zr_set_norm(zr, tnp, 1);

		break;

	case DECODER_GET_TVNORM:
		if (zr->zr_dec_mode == VIDEO_MODE_ERROR)
			error = ENXIO;
		else
			*(struct tvnorm *)data = *zr->zr_dec_norm;
		break;

	case DECODER_SETGEO:
		dg = (struct decoder_geomet *)data;

		if (zr->zr_maskalloc) {
			error = EACCES;
			break;
		}

		if ((error = zr_goto_state(zr, ZR_WINSET)))
			break;

		zr->zr_buffer.format = dg->oformat;
		if (zr->zr_buffer.width != dg->columns || 
		    zr->zr_buffer.height != dg->rows)
			zr->zr_maskvalid = 0;
		zr->zr_buffer.width = dg->columns;
		zr->zr_buffer.height = dg->rows;

		break;

	case DECODER_GETGEO:
		dg = (struct decoder_geomet *)data;

		if (zr->zr_winset) {
			dg->oformat = zr->zr_buffer.format;
			dg->columns = zr->zr_buffer.width;
			dg->rows = zr->zr_buffer.height;
		} else
			error = ENXIO;

		break;

	case DECODER_SVIDEO:
		dv = (struct decoder_vram *)data;

		ZRSC_PRINTF(zr, "SVIDEO is a BAD interface, it lets the user do REALLY stupid things by passing the physical address in.\n");

		if ((error = zr_goto_state(zr, ZR_BUFSET)))
			break;

		zr->zr_physaddr = dv->addr;
		zr->zr_buffer.addr = NULL;
		zr->zr_buffer.vidwidth = dv->width;
		zr->zr_buffer.len = dv->ramsize;
		zr->zr_buffer.xstart = zr->zr_buffer.ystart = 0;

		break;

	case DECODER_GVIDEO:
		dv = (struct decoder_vram *)data;

		if (zr->zr_bufset) {
			dv->addr = zr->zr_physaddr;
			dv->width = zr->zr_buffer.vidwidth;
			dv->ramsize = zr->zr_buffer.len;
		} else
			error = ENXIO;

		break;
		
	case DECODER_SVIDBUF:
		db = (struct decoder_buffer *)data;

		if (zr->zr_maskalloc) {
			error = EACCES;
			break;
		}

		if (proc != zr->zr_proc) {
			ZRSC_PRINTF(zr,
			    "proc doesn't match proc that opened, proc: %p, zr->zr_proc: %p.\n", proc, zr->zr_proc);
			error = EFAULT;
			break;
		}

		if ((error = verify_contig(zr->zr_proc, db->addr, db->len,
		    &phys)))
			break;
		/*printf("verify_contig: %d, phys: 0x%x\n", error, phys);*/

		/*
		 * XXX - we really need to do the both at the same time.
		 */
		i = zr->zr_bufset;
		if ((error = zr_goto_state(zr, ZR_BUFSET)))
			break;

		if ((error = zr_goto_state(zr, ZR_WINSET))) {
			zr->zr_bufset = i;
			break;
		}

		zr->zr_maskvalid = 0;
		zr->zr_buffer = *db;
		zr->zr_physaddr = phys;

		break;

	case DECODER_GVIDBUF:
		if (zr->zr_bufset && zr->zr_winset)
			*((struct decoder_buffer *)data) = zr->zr_buffer;
		else
			error = ENXIO;

		break;

	case DECODER_CAPTURE:
		i = *(int *)data;

		if (zr->zr_card == NULL || zr->zr_decoder == NULL) {
			error = ENXIO;
			break;
		}

		switch (i) {
		case DECODER_CAP_CONTINOUS:
			if ((error = zr_goto_state(zr, ZR_CAPTURE_START))) {
				ZRSC_PRINTF(zr, "goto state failed\n");
				break;
			}

			if ((error = zr_set_vfe(zr))) {
				ZR_DUMB_GOTO(zr, ZR_CAPTURE_STOP);
				break;
			}

			VIDEOCODE_ENABLE_OUTPUT(zr->zr_decoder);
			zr_gpio(zr, zr->zr_card->gpio_setvidbusdir, 0);
			zr_gpio(zr, zr->zr_card->gpio_setvidbusen, 0);
			zr_gpio(zr, zr->zr_card->gpio_freq, 0);
			ZR_SET_ASR(zr, ZR_ICR_REG, icr, GIRQ1En, 1);
			ZR_SET_ASR(zr, ZR_VDCR_REG, vdcr, VidEn, 1);

			/*ZRSC_PRINTF(zr, "Enabling Video\n");
			ZRSC_PRINTF(zr, "VFE Registers:\n");
			ZR_DUMPREGS(zr, vfe_regs);
			ZRSC_PRINTF(zr, "Control Registers:\n");
			ZR_DUMPREGS(zr, control_regs);*/

			break;

		case DECODER_CAP_STOP_CONT:
			if ((error = zr_goto_state(zr, ZR_CAPTURE_STOP))) {
				ZRSC_PRINTF(zr, "goto state failed\n");
				break;
			}

			zr_gpio(zr, zr->zr_card->gpio_setvidbusen, 1);
			ZR_SET_ASR(zr, ZR_ICR_REG, icr, GIRQ1En, 0);
			ZR_SET_ASR(zr, ZR_VDCR_REG, vdcr, VidEn, 0);
			VIDEOCODE_DISABLE_OUTPUT(zr->zr_decoder);

			/*ZRSC_PRINTF(zr, "Disabling Video\n");
			ZRSC_PRINTF(zr, "VFE Registers:\n");
			ZR_DUMPREGS(zr, vfe_regs);
			ZRSC_PRINTF(zr, "Control Registers:\n");
			ZR_DUMPREGS(zr, control_regs);*/

			break;

		default:
			error = EINVAL;
		}
		break;

	case VIDEO_OVERLAY:
		omp = (struct overlay_mask *)data;

		if (!zr->zr_winset) {
			error = EACCES;
			break;
		}

		if ((omp->om_allvis && omp->om_allhide) ||
		    (omp->om_vis && omp->om_hide) ||
		    (omp->om_dealloc && !zr->zr_mask) ||
		    (!zr->zr_mask && !omp->om_allvis && !omp->om_allhide) ||
		    omp->om_cnt > VIDEO_OVERLAY_NBOXES) {
			error = EINVAL;
			break;
		}

		if (omp->om_dealloc) {
			zr_mask_free(zr);
			break;
		}

		zr->zr_maskstride = ms = (zr->zr_buffer.width + 31) / 32 * 4;

		/*ZRSC_PRINTF(zr, "ms: %d\n", ms);*/

		/* allocate the mask if we haven't already */
		if (!zr->zr_mask) {
			if ((error = zr_goto_state(zr, ZR_MASKALLOC))) {
				ZRSC_PRINTF(zr, "bleh, goto_state\n");
				break;
			}

			zr->zr_masklen = zr->zr_buffer.height * ms;

			if ((error = zr_get_mem(zr, zr->zr_bus.zb_tag,
			    &zr->zr_bus.zb_mask, zr->zr_masklen, 1))) {
				ZR_DUMB_GOTO(zr, ZR_MASKFREE);
				ZRSC_PRINTF(zr, "bleh, get_mem, error: %d\n",
				    error);
				break;
			}

			ZR_DUMB_GOTO(zr, ZR_MASK);

			zr->zr_defvis = omp->om_allvis ? 1 : 0;

			memset((caddr_t)zr->zr_bus.zb_mask.zd_vaddr,
			    zr->zr_defvis ? 0xff : 0, zr->zr_masklen);
		} else if (omp->om_allvis || omp->om_allhide)
			memset((caddr_t)zr->zr_bus.zb_mask.zd_vaddr,
			    omp->om_allvis ? 0xff : 0, zr->zr_masklen);

		zr->zr_maskvalid = 1;
		mask = (caddr_t)zr->zr_bus.zb_mask.zd_vaddr;

		/*ZRSC_PRINTF(zr, "going through boxes: %d\n",
		    (int)omp->om_cnt);*/

		for (i = 0; i < omp->om_cnt; i++) {
			ob = omp->om_boxes[i];
			if (ob.x1 > ob.x2 || ob.y1 > ob.y2) {
				error = EINVAL;
				break;
			}

			for (j = ob.y1; j <= ob.y2; j++) {
				if (ob.x1 / 8 == ob.x2 / 8) {
					if (omp->om_vis)
						mask[ob.x1 / 8 + j * ms] |=
						    head_mask[ob.x1 & 0x7] &
						    tail_mask[ob.x2 & 0x7];
					else
						mask[ob.x1 / 8 + j * ms] &=
						    ~(head_mask[ob.x1 & 0x7] &
						    tail_mask[ob.x2 & 0x7]);
				} else {
					if (omp->om_vis)
						mask[ob.x1 / 8 + j * ms] |=
						    head_mask[ob.x1 & 0x7];
					else
						mask[ob.x1 / 8 + j * ms] &=
						    ~(head_mask[ob.x1 & 0x7]);
					for (k = ob.x1 / 8 + 1; k < ob.x2 / 8;
					    k++)
						mask[k + j * ms] =
						    omp->om_vis ? 0xff : 0x00;
					if (omp->om_vis)
						mask[ob.x2 / 8 + j * ms] |=
						    tail_mask[ob.x2 & 0x7];
					else
						mask[ob.x2 / 8 + j * ms] &=
						    ~(tail_mask[ob.x2 & 0x7]);
					
				}
			}
		}
		break;

	default:
		error = EINVAL;
		break;
	}

	/*if (error)
		ZRSC_PRINTF(zr, "ioctl returning error: %d\n", error);*/

	return error;
}

/*
 * Clears all intrupts even if they weren't handled to prevent interrupt
 * storms.
 */
void
zr_core_intr(struct zr_softc *zr)
{
	asrs_t asr;
#ifdef ZR_DEBUG_ISR
	static asrs_t last;
	static int h = 0;
#endif


	asr.reg = zr_read_asrs(zr, ZR_ISR_REG);
#ifdef ZR_DEBUG_ISR
	if (!h) {
		ZRSC_PRINTF(zr, "intr\n");
		h = 1;
	}

	if (last.reg != asr.reg) {
		ZRSC_PRINTF(zr, "ISR: 0x%x\n", asr.reg);
		last = asr;
	}
#endif

	if (asr.isr.JPEGRepIRQ)
		zr->zr_irqcnt.JPEGRepIRQ++;
	if (asr.isr.CodRepIRQ)
		zr->zr_irqcnt.CodRepIRQ++;
	if (asr.isr.GIRQ0)
		zr->zr_irqcnt.GIRQ0++;
	if (asr.isr.GIRQ1)
		zr->zr_irqcnt.GIRQ1++;
	/*if (zr->zr_irqcnt.GIRQ1 && (zr->zr_irqcnt.GIRQ1 % 20) == 0)
		ZRSC_PRINTF(zr, "girq1: %d\n", zr->zr_irqcnt.GIRQ1);*/

	/* you write the bit that you want to clear! */
	zr_write_asrs(zr, ZR_ISR_REG, asr.reg);
}
