/*-
 * Copyright 2003 John-Mark Gurney
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: zr36060.c,v 2.1 2003/05/04 15:01:03 jmg Exp $
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/conf.h>
#include <sys/uio.h>
#include <sys/kernel.h>
#include <sys/ioccom.h>
#include <sys/malloc.h>
#include <sys/mutex.h>
#include <sys/signalvar.h>
#include <sys/mman.h>
#include <sys/poll.h>
#include <sys/select.h>

#include <vm/vm.h>
#include <vm/vm_kern.h>
#include <vm/pmap.h>
#include <vm/vm_extern.h>

#include <machine/bus_memio.h>	/* used by bus space */
#include <machine/bus.h>	/* used by bus space and newbus */
#include <sys/bus.h>

#include <sys/rman.h>
#include <machine/resource.h>
#include <machine/clock.h>	/* XX */

#include <dev/zr36060/zr36060_reg.h>
#include "device_if.h"

#include <dev/postoffice/postoffice.h>
#include "postoffice_if.h"

#include <dev/videocode/videosubdev.h>
#include "videosubdev_if.h"

#include <dev/videocode/ioctl_video.h>
#include <dev/videocode/videocode.h>

#define ZR60_KEY	0xbf5600
#define ZR60_SOFTC	struct zr60_softc *z60 = device_get_softc(dev)

/*
 * Data
 */
struct reg_block {
	u_int16_t reg;
	u_int16_t len;
	u_char	*data;
};

#define Z60_DECLARE_REG_BLOCK(reg, data)		\
	(struct reg_block){ (reg), sizeof((u_char *)(data)), (data) }

static MALLOC_DEFINE(M_ZR36060, "zr36060", "Zoran 36060 data");

static u_char sos_marker[] = { 0xff, 0xda, 0x00, 0x0c,
	0x03, 0x00, 0x00, 0x01, 0x11, 0x02, 0x11, 0x00, 0x3f, 0x00 };
static u_char dri_marker[] = { 0xff, 0xd0, 0x00, 0x04,
	0x00, 0x08 };
static u_char dqt_marker[] = { 0xff, 0xd8, 0x00, 0x84,
	0x00, 0x10, 0x0b, 0x0c, 0x0e, 0x0c, 0x0a, 0x10, 0x0e, 0x0d,
	0x0e, 0x12, 0x11, 0x10, 0x13, 0x18, 0x28, 0x1a, 0x18, 0x16,
	0x16, 0x18, 0x31, 0x23, 0x25, 0x1d, 0x28, 0x3a, 0x33, 0x3d,
	0x3c, 0x39, 0x33, 0x38, 0x37, 0x40, 0x48, 0x5c, 0x4e, 0x40,
	0x44, 0x57, 0x45, 0x37, 0x38, 0x50, 0x6d, 0x51, 0x57, 0x5f,
	0x62, 0x67, 0x68, 0x67, 0x3e, 0x4d, 0x71, 0x79, 0x70, 0x64,
	0x78, 0x5c, 0x65, 0x67, 0x63, 0x01, 0x11, 0x12, 0x12, 0x18,
	0x15, 0x18, 0x2f, 0x1a, 0x1a, 0x2f, 0x63, 0x42, 0x38, 0x42,
	0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63,
	0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63,
	0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63,
	0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63,
	0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63 };
static u_char dht_marker[] = { 0xff, 0xc4, 0x01, 0xa2,
	0x00, 0x00, 0x01, 0x05, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x02,
	0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0A, 0x0B, 0x01,
	0x00, 0x03, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01,
	0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x02, 0x03,
	0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0A, 0x0B, 0x10, 0x00,
	0x02, 0x01, 0x03, 0x03, 0x02, 0x04, 0x03, 0x05, 0x05, 0x04,
	0x04, 0x00, 0x00, 0x01, 0x7D, 0x01, 0x02, 0x03, 0x00, 0x04,
	0x11, 0x05, 0x12, 0x21, 0x31, 0x41, 0x06, 0x13, 0x51, 0x61,
	0x07, 0x22, 0x71, 0x14, 0x32, 0x81, 0x91, 0xA1, 0x08, 0x23,
	0x42, 0xB1, 0xC1, 0x15, 0x52, 0xD1, 0xF0, 0x24, 0x33, 0x62,
	0x72, 0x82, 0x09, 0x0A, 0x16, 0x17, 0x18, 0x19, 0x1A, 0x25,
	0x26, 0x27, 0x28, 0x29, 0x2A, 0x34, 0x35, 0x36, 0x37, 0x38,
	0x39, 0x3A, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49, 0x4A,
	0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59, 0x5A, 0x63, 0x64,
	0x65, 0x66, 0x67, 0x68, 0x69, 0x6A, 0x73, 0x74, 0x75, 0x76,
	0x77, 0x78, 0x79, 0x7A, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88,
	0x89, 0x8A, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98, 0x99,
	0x9A, 0xA2, 0xA3, 0xA4, 0xA5, 0xA6, 0xA7, 0xA8, 0xA9, 0xAA,
	0xB2, 0xB3, 0xB4, 0xB5, 0xB6, 0xB7, 0xB8, 0xB9, 0xBA, 0xC2,
	0xC3, 0xC4, 0xC5, 0xC6, 0xC7, 0xC8, 0xC9, 0xCA, 0xD2, 0xD3,
	0xD4, 0xD5, 0xD6, 0xD7, 0xD8, 0xD9, 0xDA, 0xE1, 0xE2, 0xE3,
	0xE4, 0xE5, 0xE6, 0xE7, 0xE8, 0xE9, 0xEA, 0xF1, 0xF2, 0xF3,
	0xF4, 0xF5, 0xF6, 0xF7, 0xF8, 0xF9, 0xFA, 0x11, 0x00, 0x02,
	0x01, 0x02, 0x04, 0x04, 0x03, 0x04, 0x07, 0x05, 0x04, 0x04,
	0x00, 0x01, 0x02, 0x77, 0x00, 0x01, 0x02, 0x03, 0x11, 0x04,
	0x05, 0x21, 0x31, 0x06, 0x12, 0x41, 0x51, 0x07, 0x61, 0x71,
	0x13, 0x22, 0x32, 0x81, 0x08, 0x14, 0x42, 0x91, 0xA1, 0xB1,
	0xC1, 0x09, 0x23, 0x33, 0x52, 0xF0, 0x15, 0x62, 0x72, 0xD1,
	0x0A, 0x16, 0x24, 0x34, 0xE1, 0x25, 0xF1, 0x17, 0x18, 0x19,
	0x1A, 0x26, 0x27, 0x28, 0x29, 0x2A, 0x35, 0x36, 0x37, 0x38,
	0x39, 0x3A, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49, 0x4A,
	0x53, 0x54, 0x55, 0x56, 0x57, 0x58, 0x59, 0x5A, 0x63, 0x64,
	0x65, 0x66, 0x67, 0x68, 0x69, 0x6A, 0x73, 0x74, 0x75, 0x76,
	0x77, 0x78, 0x79, 0x7A, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87,
	0x88, 0x89, 0x8A, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98,
	0x99, 0x9A, 0xA2, 0xA3, 0xA4, 0xA5, 0xA6, 0xA7, 0xA8, 0xA9,
	0xAA, 0xB2, 0xB3, 0xB4, 0xB5, 0xB6, 0xB7, 0xB8, 0xB9, 0xBA,
	0xC2, 0xC3, 0xC4, 0xC5, 0xC6, 0xC7, 0xC8, 0xC9, 0xCA, 0xD2,
	0xD3, 0xD4, 0xD5, 0xD6, 0xD7, 0xD8, 0xD9, 0xDA, 0xE2, 0xE3,
	0xE4, 0xE5, 0xE6, 0xE7, 0xE8, 0xE9, 0xEA, 0xF2, 0xF3, 0xF4,
	0xF5, 0xF6, 0xF7, 0xF8, 0xF9, 0xFA
	};

static struct reg_block markers[] = {
	Z60_DECLARE_REG_BLOCK(0x07a, sos_marker),
	Z60_DECLARE_REG_BLOCK(0x0c0, dri_marker),
	Z60_DECLARE_REG_BLOCK(0x0d0, dqt_marker),
	Z60_DECLARE_REG_BLOCK(0x1d4, dht_marker)
};

/*
 | ZR36060
 |	Integrated JPEG Codec
 */
static int
zr36060_id_write8(device_t dev, u_int8_t id, u_int16_t reg, u_int8_t val)
{
	if (POSTOFFICE_WRITE(dev, id, 1, reg >> 8)
	    || POSTOFFICE_WRITE(dev, id, 2, reg)
	    || POSTOFFICE_WRITE(dev, id, 3, val))
		return -1;
	return 0;
}

static int
zr36060_write8(struct zr60_softc *z60, u_int16_t reg, u_int8_t val)
{
	return zr36060_id_write8(z60->z60_parent, z60->z60_id, reg, val);
}

static int
zr36060_write16(struct zr60_softc *z60, u_int16_t reg, u_int16_t val)
{
	device_t dev;
	u_int8_t id;
	int error;
	/*
	 * XXX - should really just call write8, but it will be faster
	 * to do it ourselves (don't have to rewrite hi addr bit)
	 */

	error = 0;
	if ((reg & 0xff) == 0xff) {
		/* we cross the addr boundary, just call write8 */
		if ((error = zr36060_write8(z60, reg, val >> 8)))
			return error;
		error = zr36060_write8(z60, reg + 1, val & 0xff);
	} else {
		dev = z60->z60_parent;
		id = z60->z60_id;

		if ((error = POSTOFFICE_WRITE(dev, id, 1, reg >> 8))
		    || (error = POSTOFFICE_WRITE(dev, id, 2, reg & 0xff))
		    || (error = POSTOFFICE_WRITE(dev, id, 3, val >> 8))
		    || (error = POSTOFFICE_WRITE(dev, id, 2, (reg + 1) & 0xff))
		    || (error = POSTOFFICE_WRITE(dev, id, 3, val & 0xff)))
			return error;
	}

	return error;
}

static int
zr36060_write_block(struct zr60_softc *z60, u_int16_t reg, caddr_t data,
	int len)
{
	device_t dev;
	u_int8_t id;
	int error;
	u_int8_t oh;
	union {
		struct {
			u_int8_t low;
			u_int8_t high;
		} p;
		u_int16_t reg;
	} rp;

	error = 0;
	dev = z60->z60_parent;
	id = z60->z60_id;

	for (rp.reg = reg, oh = !rp.p.high; rp.reg < reg + len;
	    rp.reg++, data++) {
		if (oh != rp.p.high && (error = POSTOFFICE_WRITE(dev, id, 1,
		    rp.p.high)))
			break;
		if ((error = POSTOFFICE_WRITE(dev, id, 2, rp.p.low))
		    || (error = POSTOFFICE_WRITE(dev, id, 3, *data)))
			break;
	}

	return error;
}

static int
zr36060_id_read8(device_t dev, u_int8_t id, u_int16_t reg, u_int8_t *retp)
{
	u_int8_t ret;

	if (retp == NULL)
		retp = &ret;

	/* if the msb don't change we don't need to do the first write */
	if(POSTOFFICE_WRITE(dev, id, 1, (reg >> 8) & 0x3)
	    || POSTOFFICE_WRITE(dev, id, 2, reg & 0xff))
		return -1;
	if (POSTOFFICE_READ(dev, id, 3, retp))
		return -1;

	return *retp;
}

static int
zr36060_read8(struct zr60_softc *z60, u_int reg, u_int8_t *ret)
{
	return zr36060_id_read8(z60->z60_parent, z60->z60_id, reg, ret);
}

static int
zr36060_read16(struct zr60_softc *z60, u_int16_t reg, u_int16_t *ret)
{
	device_t dev;
	u_int8_t id;
	u_int8_t reth, retl;
	int error;

	/*
	 * XXX - should really just call read8, but it will be faster
	 * to do it ourselves (don't have to rewrite hi addr bit)
	 */

	error = 0;
	if ((reg & 0xff) == 0xff) {
		/* we cross the addr boundary, just call read8 */
		if ((error = zr36060_read8(z60, reg, &reth)) == -1)
			return error;
		if ((error = zr36060_read8(z60, reg + 1, &retl)) == -1)
			return error;
	} else {
		dev = z60->z60_parent;
		id = z60->z60_id;

		if ((error = POSTOFFICE_WRITE(dev, id, 1, reg >> 8))
		    || (error = POSTOFFICE_WRITE(dev, id, 2, reg & 0xff))
		    || (error = POSTOFFICE_READ(dev, id, 3, &reth)) == -1
		    || (error = POSTOFFICE_WRITE(dev, id, 2, (reg + 1) & 0xff))
		    || (error = POSTOFFICE_READ(dev, id, 3, &retl)) == -1)
			return error;
	}

	if (ret != NULL)
		*ret = reth << 8 | retl;

	return reth << 8 | retl;
}

static int
zr36060_read_block(struct zr60_softc *z60, u_int16_t reg, caddr_t data,
	int len)
{
	device_t dev;
	u_int8_t id;
	int error;
	u_int8_t oh;
	union {
		struct {
			u_int8_t low;
			u_int8_t high;
		} p;
		u_int16_t reg;
	} rp;

	error = 0;
	dev = z60->z60_parent;
	id = z60->z60_id;

	for (rp.reg = reg, oh = !rp.p.high; rp.reg < reg + len;
	    rp.reg++, data++) {
		if (oh != rp.p.high && (error = POSTOFFICE_WRITE(dev, id, 1,
		    rp.p.high)))
			break;
		if ((error = POSTOFFICE_WRITE(dev, id, 2, rp.p.low))
		    || (error = POSTOFFICE_READ(dev, id, 3, data)) == -1)
			break;
	}

	return error;
}

static void
zr36060_find_coders(struct zr60_softc *z60)
{
	int i;

	if (z60->z60_decoder == NULL) {
		i = 1;
		videocode_findtype(z60->z60_parent,
		    VIDEO_DECODER, &z60->z60_decoder, &i);
		if (i != 1)
			z60->z60_decoder = NULL;
		else
			device_busy(z60->z60_decoder);
	}
}

static void
zr36060_release_coders(struct zr60_softc *z60)
{
	if (z60->z60_decoder != NULL) {
		device_unbusy(z60->z60_decoder);
		z60->z60_decoder = NULL;
	}
}

static int
zr36060_init_chip(struct zr60_softc *z60)
{
	int i;
	struct reg_block *rbp;

	for (i = 0; i < sizeof markers / sizeof *markers; i++) {
		rbp = &markers[i];
		if (zr36060_write_block(z60, rbp->reg, rbp->data, rbp->len))
			return EIO;
	}

	if (zr36060_write8(z60, 0x002, (1 << 0))	/* CIR */
	    || zr36060_write8(z60, 0x004, 0)		/* Res */
	    || zr36060_write8(z60, 0x006, 0)		/* MER */
	    || zr36060_write16(z60, 0x024, 0))		/* TCR */
		return EIO;

	return 0;
}

/*
 * XXX - need to get the GPIO pin information from the parent.
 */
static int
zr36060_dev_sleep(device_t dev, u_int8_t gpio, u_int8_t val)
{
     POSTOFFICE_GPIO(dev, gpio, !val);
     if(val == 0)
	  DELAY(500);
     else
	  DELAY(2);
     return 0;
}

static int
zr36060_sleep(struct zr60_softc *z60, u_int8_t val)
{
	return zr36060_dev_sleep(z60->z60_parent, z60->z60_slp, val);
}

static int 
zr36060_dev_reset(device_t dev, u_int8_t slp, u_int8_t rst)
{
     zr36060_dev_sleep(dev, slp, 1);
     POSTOFFICE_GPIO(dev, rst, 0);
     DELAY(2);
     POSTOFFICE_GPIO(dev, rst, 1);
     DELAY(2);
     zr36060_dev_sleep(dev, slp, 0);

     return 0;
}

static int 
zr36060_reset(struct zr60_softc *z60)
{
	return zr36060_dev_reset(z60->z60_parent, z60->z60_slp, z60->z60_rst);
}

static void
zr36060_set_norm(struct zr60_softc *z60, const struct tvnorm *tnp, int copy)
{
	struct tvnorm *alloc;

	alloc = NULL;
	if (copy)
		alloc = malloc(sizeof *alloc, M_ZR36060, M_WAITOK);

	mtx_lock(&z60->z60_lock);

	if (copy) {
		if (!z60->z60_norm_allocated) {
			z60->z60_norm_allocated = 1;
		} else {
			free(alloc, M_ZR36060);
			alloc = (struct tvnorm *)(uintptr_t)z60->z60_norm;
		}

		*alloc = *tnp;
		z60->z60_norm = alloc;
	} else {
		if (z60->z60_norm_allocated)
			free((void *)(uintptr_t)z60->z60_norm, M_ZR36060);
		z60->z60_norm = tnp;
	}

	mtx_unlock(&z60->z60_lock);
}

static int
zr36060_probe(device_t dev)
{
	ZR60_SOFTC;
	int error;
	device_t parent;
	uintptr_t i, j, k;

	z60->z60_parent = parent = device_get_parent(dev);
	z60->z60_self = dev;
	if ((error = BUS_READ_IVAR(z60->z60_parent, dev, PO_IVAR_ID,
	    &i)) || (error = BUS_READ_IVAR(z60->z60_parent, dev,
	    PO_IVAR_SLEEP, &j)) ||
	    (error = BUS_READ_IVAR(z60->z60_parent, dev, PO_IVAR_RESET,
	    &k)))
		return error;

	z60->z60_id = i;
	z60->z60_slp = j;
	z60->z60_rst = k;

	POSTOFFICE_SET_DURREC(z60->z60_parent, z60->z60_id, 12, 3);
	zr36060_reset(z60);

	if (zr36060_id_read8(parent, z60->z60_id, 0x22, NULL) != 0x33)
		return ENXIO;

	device_set_desc(dev, "ZR36060 MJPEG Codec");

	return 0;
}

static int
zr36060_attach(device_t dev)
{
	int error;
	ZR60_SOFTC;

	if (!mtx_initialized(&z60->z60_lock))
		mtx_init(&z60->z60_lock, device_get_nameunit(dev), NULL,
		    MTX_DEF);

	/* this should detect the parent, and behave accordingly */
	/* for now assume the ZR36067-LV */
	/*POSTOFFICE_SET_DURREC(z60->z60_parent, z60->z60_id, 12, 3);*/

	if (zr36060_read8(z60, 0x23, &z60->z60_rev) == -1)
		z60->z60_rev = -1;

	device_printf(dev, "ZR36060 (codec rev %d)\n", (int)z60->z60_rev);

	if ((error = zr36060_init_chip(z60)))
		device_printf(dev, "init_chip error: %d\n", error);

	return 0;
}

static int
zr36060_detach(device_t dev)
{
	ZR60_SOFTC;

	zr36060_sleep(z60, 1);

	z60->z60_self = NULL;
	z60->z60_parent = NULL;

	mtx_destroy(&z60->z60_lock);

	return 0;
}

static int
zr36060_shutdown(device_t dev)
{
	ZR60_SOFTC;

	zr36060_sleep(z60, 1);

	return 0;
}

static int
zr36060_open(device_t dev)
{
	device_busy(dev);

	return 0;
}

static int
zr36060_close(device_t dev)
{
	ZR60_SOFTC;

	zr36060_release_coders(z60);

	device_unbusy(dev);

	return 0;
}

static int
zr36060_ioctl(device_t dev, u_long cmd, caddr_t data, int fflag,
        struct thread *td)
{
	ZR60_SOFTC;
	int error;
	struct jpeg_marker *jmp;
	struct tvnorm *tnp;
	u_int16_t mark, len;

	error = 0;
	switch (cmd) {
	case MJPEG_SET_TVNORM:
		if (z60->z60_norm == NULL) {
			IFCODERINVAL(z60, decoder) {
				error = ENXIO;
				break;
			}
			z60->z60_norm = VIDEOCODE_GET_TVNORM(z60->z60_decoder);
			if (z60->z60_norm == NULL) {
				error = ENXIO;
				break;
			}
		}
		tnp = (struct tvnorm *)data;

		if (tnp->Wt != z60->z60_norm->Wt ||
		    tnp->Ht != z60->z60_norm->Ht ||
		    tnp->HSyncStart != z60->z60_norm->HSyncStart) {
			error = EINVAL;
			break;
		}

		zr36060_set_norm(z60, tnp, 1);

		break;

	case MJPEG_GET_TVNORM:
		if (z60->z60_norm == NULL) {
			IFCODERINVAL(z60, decoder) {
				error = ENXIO;
				break;
			}
			z60->z60_norm = VIDEOCODE_GET_TVNORM(z60->z60_decoder);
			if (z60->z60_norm == NULL) {
				error = ENXIO;
				break;
			}
		}
		*(struct tvnorm *)data = *z60->z60_norm;

		break;

	case MJPEG_SET_APP:
		jmp = (struct jpeg_marker *)data;

		if (jmp->jm_id < 0xE0 || jmp->jm_id > 0xEf
		    || jmp->jm_len > 62) {
			error = EINVAL;
			break;
		}

		if (zr36060_write16(z60, 0x380, 0xFF << 8 | jmp->jm_id)
		    || zr36060_write16(z60, 0x382, jmp->jm_len)
		    || zr36060_write_block(z60, 0x384, jmp->jm_data,
		    jmp->jm_len))
			error = EIO;

		break;

	case MJPEG_GET_APP:
		jmp = (struct jpeg_marker *)data;

		if (zr36060_read16(z60, 0x380, &mark) == -1
		    || zr36060_read16(z60, 0x382, &len) == -1) {
			error = EIO;
			break;
		}

		if (len > 62)
			len = 62;
		jmp->jm_len = len;

		if (mark >> 8 != 0xff || (mark & 0xff) < 0xE0
		    || (mark & 0xff) > 0xEF) {
			error = EINVAL;
			break;
		}

		jmp->jm_id = mark & 0xff;
		if (zr36060_read_block(z60, 0x384, jmp->jm_data, len) == -1) {
			error = EIO;
			break;
		}

		break;

	case MJPEG_SET_COM:
		jmp = (struct jpeg_marker *)data;

		if (jmp->jm_id < 0xE0 || (jmp->jm_id > 0xEF
		    && jmp->jm_id != 0xFE) || jmp->jm_len > 62) {
			error = EINVAL;
			break;
		}

		if (zr36060_write16(z60, 0x3C0, 0xFF << 8 | jmp->jm_id)
		    || zr36060_write16(z60, 0x3C2, jmp->jm_len)
		    || zr36060_write_block(z60, 0x3C4, jmp->jm_data,
		    jmp->jm_len))
			error = EIO;

		break;

	case MJPEG_GET_COM:
		jmp = (struct jpeg_marker *)data;

		if (zr36060_read16(z60, 0x3C0, &mark) == -1
		    || zr36060_read16(z60, 0x3C2, &len) == -1) {
			error = EIO;
			break;
		}

		if (len > 62)
			len = 62;
		jmp->jm_len = len;

		if (mark >> 8 != 0xff || (mark & 0xff) < 0xE0
		    || ((mark & 0xff) > 0xEF && (mark & 0xff) != 0xFE)) {
			error = EINVAL;
			break;
		}

		jmp->jm_id = mark & 0xff;
		if (zr36060_read_block(z60, 0x3C4, jmp->jm_data, len) == -1) {
			error = EIO;
			break;
		}

		break;

	default:
		error = EINVAL;
		break;
	}

	return error;
}

static void
zr36060_intr(device_t dev)
{
	device_printf(dev, "intr called\n");
}

static device_method_t zr36060_methods[] = {
	/* device interface */
	DEVMETHOD(device_probe,		zr36060_probe),
	DEVMETHOD(device_attach,	zr36060_attach),
	DEVMETHOD(device_detach,	zr36060_detach),
	DEVMETHOD(device_shutdown,	zr36060_shutdown),

	/* video subdev interface */
	DEVMETHOD(videosubdev_open,	zr36060_open),
	DEVMETHOD(videosubdev_close,	zr36060_close),
	DEVMETHOD(videosubdev_ioctl,	zr36060_ioctl),
	DEVMETHOD(videosubdev_intr,	zr36060_intr),

	{ 0, 0}
	};

static driver_t zr36060_driver = {
        "zr36060",
        zr36060_methods,
        sizeof(struct zr60_softc)
};

static devclass_t zr36060_devclass;

DRIVER_MODULE(zr36060, zr, zr36060_driver, zr36060_devclass, 0, 0);
MODULE_DEPEND(zr36060, postoffice, POSTOFFICE_MINVER, POSTOFFICE_PREFVER,
	POSTOFFICE_MAXVER);
MODULE_DEPEND(zr, videocode, VIDEOCODE_MINVER, VIDEOCODE_PREFVER,
	VIDEOCODE_MAXVER);
MODULE_DEPEND(zr, videosubdev, VIDEOSUBDEV_MINVER, VIDEOSUBDEV_PREFVER,
	VIDEOSUBDEV_MAXVER);

MODULE_VERSION(zr36060, 1);
